# -*- coding: utf-8 -*-
from typing import Any, List, Deque, Union, Generic, TypeVar
from collections import deque

from eventbus.domain.entity import TimestampedVersionedEntity, DomainEntity, TDomainEvent
from eventbus.domain.events import DomainEvent

TAggregate = TypeVar("TAggregate", bound="BaseAggregateRoot")
TAggregateEvent = TypeVar("TAggregateEvent", bound="BaseAggregateRoot.Event")


class BaseAggregateRoot(TimestampedVersionedEntity, Generic[TAggregateEvent]):
    """
    Root entity for an aggregate in a domain driven design.
    """

    class Event(TimestampedVersionedEntity.Event[TAggregate]):
        """Supertype for base aggregate root events."""

    class Created(TimestampedVersionedEntity.Created[TAggregate], Event[TAggregate]):
        """Triggered when an aggregate root is created."""

    class AttributeChanged(Event[TAggregate], TimestampedVersionedEntity.AttributeChanged[TAggregate]):
        """Triggered when an aggregate root attribute is changed."""

    class Discarded(Event[TAggregate], TimestampedVersionedEntity.Discarded[TAggregate]):
        """Triggered when an aggregate root is discarded."""

    def __init__(self, **kwargs: Any) -> None:
        super(BaseAggregateRoot, self).__init__(**kwargs)
        self.__pending_events__: Deque[DomainEntity.Event] = deque()

    async def __publish__(self, events: Union[TDomainEvent, List[TDomainEvent]]) -> None:
        """
        Defers publishing event(s) to subscribers, by adding
        event to internal collection of pending events.
        """
        if isinstance(events, DomainEvent):
            events = [events]

        self.__pending_events__.extend(events)

    async def __save__(self) -> None:
        """
        Publishes all pending events to subscribers.
        """
        batch_of_events = self.__batch_pending_events__()
        if batch_of_events:
            await self._event_bus.publish(batch_of_events)
            # Don't catch exception and put the events back on the queue.
            # Losing them here is consistent with the behaviour of DomainEntity
            # when an event cannot be stored: the event is effectively lost, the
            # state of the entity must be reset, and the operation repeated.
            # In case of an aggregate sequence conflict, developers need to
            # know what has happened since the last save, so can retry only the
            # command(s) that have caused conflict. Best to save once per command,
            # so the command can be retried cleanly. The purpose of the save
            # method is to allow many events from one command to be persisted
            # together. The save command can be used to persist all the events
            # from many commands, but in case of a failed save after several
            # commands have been executed, it is important to know which
            # commands to retry.

    def __batch_pending_events__(self) -> List[DomainEntity.Event]:
        batch_of_events: List[DomainEntity.Event] = []
        try:
            while True:
                batch_of_events.append(self.__pending_events__.popleft())
        except IndexError:
            pass
        return batch_of_events



# -*- coding: utf-8 -*-
class EventSourcingError(Exception):
    """Base eventsourcing exception."""


class TopicResolutionError(EventSourcingError):
    """Raised when unable to resolve a topic to a Python class."""


class ConsistencyError(EventSourcingError):
    """Raised when applying an event stream to a versioned entity."""


class MismatchedOriginatorError(ConsistencyError):
    """Raised when applying an event to an inappropriate object."""


class OriginatorIDError(MismatchedOriginatorError):
    """Raised when applying an event to the wrong entity or aggregate."""


class OriginatorVersionError(MismatchedOriginatorError):
    """Raised when applying an event to the wrong version of an entity or aggregate."""


class DataIntegrityError(ValueError, EventSourcingError):
    """ Raised when a sequenced item is damaged (hash doesn't match data)"""


class EntityIsDiscarded(AssertionError):
    """Raised when access to a recently discarded entity object is attempted."""


class ProgrammingError(EventSourcingError):
    """Raised when programming errors are encountered."""


class EncoderTypeError(TypeError):
    pass



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import List, Tuple, Type, Union, Optional

from eventbus.domain.whitehead import TEvent


class AbstractEventHandler(ABC):
    ASTERISK = "*"

    @property
    @abstractmethod
    def event_type(self) -> Union[str, Type, Tuple]:
        pass

    def filter(self, events: List[TEvent]):
        # When self.event_type == self.ASTERISK -> return all available events
        return list(filter(lambda elm: self.event_type == self.ASTERISK or isinstance(elm, self.event_type), events))

    def predicate(self, events) -> bool:
        return len(self.filter(events)) > 0

    @abstractmethod
    async def handler(self, events: List[TEvent]):
        pass


class AbstractEventBus(ABC):
    @abstractmethod
    def subscribe(self, handler: AbstractEventHandler) -> None:
        pass

    @abstractmethod
    def unsubscribe(self, handler: AbstractEventHandler) -> None:
        pass

    @abstractmethod
    async def publish(self, events: List[TEvent]) -> None:
        pass



# -*- coding: utf-8 -*-
from typing import List, Optional, Union, Type, Tuple

import pytest

from eventbus.application.eventbus import EventBus
from eventbus.domain.eventbus import AbstractEventHandler
from eventbus.domain.events import DomainEvent
from eventbus.domain.whitehead import TEvent, T
from eventbus.example.entity import Example


@pytest.mark.asyncio
async def test_example_entity():
    received_events = []

    class SomeEvent(DomainEvent):
        pass

    class NonExistentEventsTestHandler(AbstractEventHandler):
        @property
        def event_type(self) -> Optional[Union[str, Type, Tuple]]:
            return SomeEvent

        async def handler(self, events: List[TEvent]):
            received_events.extend(events)

    class AllEventsTestHandler(AbstractEventHandler):
        @property
        def event_type(self) -> Union[str, Type, Tuple]:
            return self.ASTERISK

        async def handler(self, events: List[TEvent]):
            received_events.extend(events)

    handler1 = AllEventsTestHandler()
    handler2 = NonExistentEventsTestHandler()

    bus = EventBus(handlers=[
        handler1, handler2
    ])

    # Same instance of handler will be skipped
    bus.subscribe(handler1)
    # No such events thus skipping as well
    bus.subscribe(handler2)

    # "Example.Created" has been added to a pending list (Domain Root Entity)
    example = await Example.__create__(event_bus=bus, first_name="First", last_name="Last", age=56)

    assert len(example.__pending_events__) == 1
    assert len(received_events) == 0

    for value in [10, 20, 30]:
        # "ExampleInternal.Created" has been published (Domain Entity)
        # "Example.ExampleInternalAdded" bas been added to a pending list (Domain Root Entity)
        await example.add(value)

    # 3x "ExampleInternal.Created"
    assert len(received_events) == 3

    # 1x "Example.Created" + 3x "Example.ExampleInternalAdded"
    assert len(example.__pending_events__) == 4

    # 1x "Example.Created" has been published
    await example.__save__()

    assert example.summ == 60

    assert len(example.__pending_events__) == 0

    # 3x "ExampleInternal.Created" + 1x "Example.Created" + 3x "ExampleInternal.Created"
    assert len(received_events) == 7



# -*- coding: utf-8 -*-
import os
import hashlib

from eventbus.util.transcoding import ObjectJSONEncoder

SALT_FOR_DATA_INTEGRITY = os.getenv("SALT_FOR_DATA_INTEGRITY", "")


def hash_object(json_encoder: ObjectJSONEncoder, obj: dict) -> str:
    """
    Calculates SHA-256 hash of JSON encoded 'obj'.
    """
    s = json_encoder.encode((obj, SALT_FOR_DATA_INTEGRITY))
    return hashlib.sha256(s).hexdigest()



# -*- coding: utf-8 -*-
from uuid import UUID
from typing import Any, Generic, Optional
from decimal import Decimal

from eventbus.domain.versioning import Upcastable
from eventbus.domain.whitehead import ActualOccasion, TEntity
from eventbus.util.hashing import hash_object
from eventbus.util.time import decimaltimestamp
from eventbus.util.topic import get_topic
from eventbus.util.transcoding import ObjectJSONEncoder


class DomainEvent(Upcastable, ActualOccasion, Generic[TEntity]):
    """
    Base class for domain model events.

    Implements methods to make instances read-only,
    comparable for equality in Python, and have
    recognisable representations.Custom
    To make domain events hashable, this class also
    implements a method to create a cryptographic hash
    of the state of the event.
    """

    __json_encoder_v2__ = ObjectJSONEncoder(sort_keys=True)
    __notifiable__ = True

    def __init__(self, **kwargs: Any):
        """
        Initialises event attribute values directly from constructor kwargs.
        """
        super().__init__()
        self.__dict__.update(kwargs)

    def __repr__(self) -> str:
        """
        Creates a string representing the type and attribute values of the event.

        :rtype: str
        """
        sorted_items = tuple(sorted(self.__dict__.items()))
        args_strings = ("{0}={1!r}".format(*item) for item in sorted_items)
        args_string = ", ".join(args_strings)
        return "{0}({1})".format(self.__class__.__qualname__, args_string)

    def __mutate__(self, obj: Optional[TEntity]) -> Optional[TEntity]:
        """
        Updates 'obj' with values from 'self'.

        Calls the 'mutate()' method.

        Can be extended, but subclasses must call super
        and return an object to their caller.

        :param obj: object (normally a domain entity) to be mutated
        :return: mutated object
        """
        if obj is not None:
            self.mutate(obj)
        return obj

    def mutate(self, obj: TEntity) -> None:
        """
        Updates ("mutates") given 'obj'.

        Intended to be overridden by subclasses, as the most
        concise way of coding a default projection of the event
        (for example into the state of a domain entity).

        The advantage of implementing a default projection
        using this method rather than __mutate__ is that you
        don't need to call super or return a value.

        :param obj: domain entity to be mutated
        """

    def __setattr__(self, key: Any, value: Any) -> None:
        """
        Inhibits event attributes from being updated by assignment.
        """
        raise AttributeError("DomainEvent attributes are read-only")

    def __eq__(self, other: object) -> bool:
        """
        Tests for equality of two event objects.

        :rtype: bool
        """
        return isinstance(other, DomainEvent) and self.__hash__() == other.__hash__()

    # Todo: Do we need this in Python 3?
    def __ne__(self, other: object) -> bool:
        """
        Negates the equality test.

        :rtype: bool
        """
        return not (self == other)

    def __hash__(self) -> int:
        """
        Computes a Python integer hash for an event.

        Supports Python equality and inequality comparisons.

        :return: Python integer hash
        :rtype: int
        """
        attrs = self.__dict__.copy()

        # Involve the topic in the hash, so that different types
        # with same attribute values have different hash values.
        attrs["__event_topic__"] = get_topic(type(self))

        # Calculate the cryptographic hash of the event.
        sha256_hash = self.__hash_object_v2__(attrs)

        # Return the Python hash of the cryptographic hash.
        return hash(sha256_hash)

    @classmethod
    def __hash_object_v2__(cls, obj: dict) -> str:
        """
        Calculates SHA-256 hash of JSON encoded 'obj'.

        :param obj: Object to be hashed.
        :return: SHA-256 as hexadecimal string.
        :rtype: str
        """
        return hash_object(cls.__json_encoder_v2__, obj)


class EventWithOriginatorID(DomainEvent[TEntity]):
    """
    For events that have an originator ID.
    """

    def __init__(self, originator_id: UUID, **kwargs: Any):
        super(EventWithOriginatorID, self).__init__(originator_id=originator_id, **kwargs)

    @property
    def originator_id(self) -> UUID:
        """
        Originator ID is the identity of the object
        that originated this event.

        :return: A UUID representing the identity of the originator.
        :rtype: UUID
        """
        return self.__dict__["originator_id"]


class EventWithTimestamp(DomainEvent[TEntity]):
    """
    For events that have a timestamp value.
    """

    def __init__(self, timestamp: Optional[Decimal] = None, **kwargs: Any):
        kwargs["timestamp"] = timestamp or decimaltimestamp()
        super(EventWithTimestamp, self).__init__(**kwargs)

    @property
    def timestamp(self) -> Decimal:
        """
        A UNIX timestamp as a Decimal object.
        """
        return self.__dict__["timestamp"]


class EventWithOriginatorVersion(DomainEvent[TEntity]):
    """
    For events that have an originator version number.
    """

    def __init__(self, originator_version: int, **kwargs: Any):
        if not isinstance(originator_version, int):
            raise TypeError("Version must be an integer: {0}".format(originator_version))
        super(EventWithOriginatorVersion, self).__init__(originator_version=originator_version, **kwargs)

    @property
    def originator_version(self) -> int:
        """
        Originator version is the version of the object
        that originated this event.

        :return: A integer representing the version of the originator.
        """
        return self.__dict__["originator_version"]


class CreatedEvent(DomainEvent[TEntity]):
    """
    Happens when something is created.
    """


class AttributeChangedEvent(DomainEvent[TEntity]):
    """
    Happens when the value of an attribute changes.
    """

    @property
    def name(self) -> str:
        return self.__dict__["name"]

    @property
    def value(self) -> Any:
        return self.__dict__["value"]


class DiscardedEvent(DomainEvent[TEntity]):
    """
    Happens when something is discarded.
    """


class LoggedEvent(DomainEvent[TEntity]):
    """
    Happens when something is logged.
    """



from typing import Dict, Type

from eventbus.domain.events import DomainEvent


def subclassevents(cls: type) -> type:  # noqa
    """
    Decorator that avoids "boilerplate" subclassing of domain events.

    For example, with this decorator you can do this:

    .. code::

        @subclassevents
        class Example(AggregateRoot):
            class SomethingHappened(DomainEvent): pass

    rather than this:

    .. code::

        class Example(AggregateRoot):
            class Event(AggregateRoot.Event): pass
            class Created(Event, AggregateRoot.Created): pass
            class Discarded(Event, AggregateRoot.Discarded): pass
            class AttributeChanged(Event, AggregateRoot.AttributeChanged): pass
            class SomethingHappened(Event): pass

    You can apply this to a tree of domain event classes by defining
    the base class with attribute 'subclassevents = True'.
    """

    bases_event_attrs = []
    super_event_class_names = set()
    for base_cls in cls.__bases__:
        base_event_attrs: Dict[str, Type[DomainEvent]] = {}
        bases_event_attrs.append(base_event_attrs)
        for base_attr_name in dir(base_cls):
            base_attr = getattr(base_cls, base_attr_name)
            if isinstance(base_attr, type) and issubclass(base_attr, DomainEvent):
                base_event_attrs[base_attr_name] = base_attr
                if base_attr_name != "Event":
                    super_event_class_names.add(base_attr_name)

    # Define base Event subclass including super Event classes.
    if "Event" in cls.__dict__:
        event_event_subclass = cls.__dict__["Event"]
    else:
        base_event_classes = []
        for base_event_attrs in bases_event_attrs:
            try:
                event_cls = base_event_attrs["Event"]
            except KeyError:
                pass
            else:
                base_event_classes.append(event_cls)

        event_event_subclass = type(
            "Event", tuple(base_event_classes or [DomainEvent]), {"__qualname__": cls.__name__ + ".Event"},
        )
        event_event_subclass.__module__ = cls.__module__
        setattr(cls, "Event", event_event_subclass)  # noqa

    # Define subclasses for super event classes, including Event subclass as base.
    for super_event_class_name in super_event_class_names:

        base_event_classes = [event_event_subclass]
        if super_event_class_name in cls.__dict__:
            continue

        for base_event_attrs in bases_event_attrs:
            try:
                event_cls = base_event_attrs[super_event_class_name]
            except KeyError:
                pass
            else:
                base_event_classes.append(event_cls)
        event_subclass = type(
            super_event_class_name,
            tuple(base_event_classes),
            {"__qualname__": cls.__name__ + "." + super_event_class_name},
        )
        event_subclass.__module__ = cls.__module__
        setattr(cls, super_event_class_name, event_subclass)

    # Redefine event classes in cls.__dict__ that are not subclasses of Event.
    for cls_attr_name in cls.__dict__.keys():
        cls_attr = getattr(cls, cls_attr_name)
        if isinstance(cls_attr, type) and issubclass(cls_attr, DomainEvent):
            if not issubclass(cls_attr, event_event_subclass):
                try:
                    event_subclass = type(
                        cls_attr_name,
                        (cls_attr, event_event_subclass),
                        {
                            "__qualname__": cls_attr.__qualname__,
                            "__module__": cls_attr.__module__,
                            "__doc__": cls_attr.__doc__,
                        },
                    )
                except TypeError:
                    raise Exception(cls_attr_name, cls_attr, event_event_subclass)
                setattr(cls, cls_attr_name, event_subclass)

    return cls



# -*- coding: utf-8 -*-
from typing import Any, List, Optional

from eventbus.domain.aggregate import BaseAggregateRoot
from eventbus.domain.entity import TimestampedVersionedEntity


class ExampleInternal(TimestampedVersionedEntity):
    def __init__(self, value: int, **kwargs: Any):
        super().__init__(**kwargs)
        self.value = value

    class Event(TimestampedVersionedEntity.Event):
        pass


class Example(BaseAggregateRoot):
    def __init__(
        self,
        first_name: str,
        last_name: str,
        age: int,
        internals: Optional[List[ExampleInternal]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.first_name = first_name
        self.last_name = last_name
        self.age = age
        self.internals: List[ExampleInternal] = internals or []

    @property
    def summ(self) -> int:
        return sum(internal.value for internal in self.internals)

    class Event(BaseAggregateRoot.Event):
        pass

    class ExampleInternalAdded(Event):
        internal: ExampleInternal

        def mutate(self, obj):
            obj.internals.append(self.internal)

    async def add(self, value: int):
        internal = await ExampleInternal.__create__(event_bus=self.event_bus, value=value)
        await self.__trigger_event__(Example.ExampleInternalAdded, internal=internal)



from typing import Dict

from eventbus.domain.whitehead import Event


class Upcastable(Event):
    """
    For things that are upcastable.

    http://code.fed.wiki.org/view/wyatt-software/view/remote-database-schema-migration

    "I was sure that we could not get the schema for WyCash Plus right
    on the first try. I was familiar with Smaltalk-80's object migration
    mechanisms. I designed a version that could serve us in a commercial
    software distribution environment.

    "I chose to version each class independently and write that as a
    sequential integer in the serialized versions. Objects would be
    mutated to the current version on read. We supported all versions
    we ever had forever.

    "I recorded mutation vectors for each version to the present. These could
    add, remove and reorder fields within an object. One-off mutation methods
    handled the rare case where the vectors were not enough description.

    "We shipped migrations to our customers with each release to be performed
    on their own machines when needed without any intervention."

    Ward Cunningham
    """

    __class_version__ = 0

    def __init__(self):
        if type(self).__class_version__ > 0:
            self.__dict__["__class_version__"] = type(self).__class_version__
        super(Upcastable, self).__init__()

    @classmethod
    def __upcast_state__(cls, obj_state: Dict) -> Dict:
        """
        Upcasts obj_state from the version of the class when the object
        state was recorded, to be compatible with current version of the class.
        """
        class_version = obj_state.get("__class_version__", 0)
        while class_version < cls.__class_version__:
            obj_state = cls.__upcast__(obj_state, class_version)
            class_version += 1
            obj_state["__class_version__"] = class_version
        return obj_state

    @classmethod
    def __upcast__(cls, obj_state: Dict, class_version: int) -> Dict:
        """
        Must be overridden in domain event classes that set class
        attribute '__class_version__' to a positive value.

        This method is expected to upcast obj_state from method arg
        'class_version' to the next version, and return obj_state.

        One style of implementation is an if-else block, in which the
        conditional expressions match on the value of 'class_version'.

        The implementation of this method must support upcasting each version
        of the class from 0 to one less than the current version.
        For example: a domain class with "__class_version__ = 1" will
        need to support upcasting "class_version == 0" to version 1; and a
        domain class with "__class_version__ = 2" will need to support both
        upcasting "class_version == 0" to version 1, and also upcasting
        "class_version == 1" to version 2.

        To support backward compatibility, the recorded state of old
        versions of an event class will need to be upcast to work with new
        versions of the software. Commonly, this involves supplying
        default values for attributes missing on old versions of events,
        after an event class has been changed by adding a new attribute.

        In a situation where a new version of an event class needs to be used
        by existing software, it would be necessary also to support forward
        compatibility. Examples of this situation include the situation
        of deploying with rolling updates, and the situation where consumers
        would receive and attempt to handle the new version of the event but
        cannot be updated before the new event class version is deployed.

        To support forward compatibility, it is necessary to restrict changes
        to be merely additive: either adding new attributes to an event, or
        adding new behaviours to the type of an existing attribute. Event
        attributes shouldn't be removed (or renamed), existing attributes
        should not have aspects of their behaviour removed, and the semantics
        of an existing attribute should not change. If the type of an attribute
        value is changed, the new type should substitutable for the old type.
        The old software will then simply ignore new attributes, and it will
        simply ignore new aspects of new types of value.

        For example, the type of an attribute value can be changed to use a
        subtype of the previous attribute value type. The possible range of a
        value can be reduced when changing the type of an attribute value, since
        all the values of the new type will be  supported by the old software.

        Increasing the possible range of a value will introduce values that
        cannot be used by the old software, and changing the type of an
        attribute value to a supertype will remove behaviour that the old
        software may depend on.

        Of course, if the old software doesn't in fact depend on aspects
        that are removed, then those aspects can be removed without actually
        breaking anything.

        If backward and forward compatibility is required, and it is felt
        that an event class needs to be changed that would cause an attribute
        to be removed or the type of an attribute to be more general, or the
        range of values of an attribute to increase, then according to Greg
        Young's book 'Versioning in an Event Sourced System' it is better to
        to add a new event type. However, if the old software would break
        because this new event type is not supported, then supporting
        forward compatibility would be elusive.

        In summary, supporting forward compatibility restricts model changes to
        adding attributes to existing model classes (which implies the versioned
        event class' upcast method will supply default values when upcasting
        the state of old versions of the event).
        """
        return obj_state



# -*- coding: utf-8 -*-
import time
import datetime
from uuid import UUID
from typing import Union, Optional
from decimal import Decimal


def decimaltimestamp_from_uuid(value: UUID) -> Decimal:
    """
    Return a floating point unix timestamp from UUID value.

    :param value:
    :return: Unix timestamp in seconds, with microsecond precision.
    :rtype: Decimal
    """
    return decimaltimestamp(timestamp_long_from_uuid(value) / 1e7)


def timestamp_long_from_uuid(value: UUID) -> int:
    """
    Returns an integer value representing a unix timestamp in tenths of microseconds.

    :param value:
    :return: Unix timestamp integer in tenths of microseconds.
    :rtype: int
    """
    if isinstance(value, str):
        value = UUID(value)

    return value.time - 0x01B21DD213814000


def decimaltimestamp(t: Optional[float] = None) -> Decimal:
    """
    A UNIX timestamp as a Decimal object (exact number type).

    Returns current time when called without args, otherwise
    converts given floating point number ``t`` to a Decimal
    with 9 decimal places.

    :param t: Floating point UNIX timestamp ("seconds since epoch").
    :return: A Decimal with 6 decimal places, representing the
            given floating point or the value returned by time.time().
    :rtype: Decimal
    """
    t = time.time() if t is None else t
    return Decimal("{0:.6f}".format(t))


def datetime_from_timestamp(t: Union[Decimal, float]) -> datetime.datetime:
    """
    Returns naive UTC datetime from decimal UNIX
    timestamps such as time.time().

    :param t: timestamp, either Decimal or float
    :return: datetime.datetime object
    """
    return datetime.datetime.fromtimestamp(float(t), tz=datetime.timezone.utc)


def timestamp_from_datetime(d: datetime.datetime) -> Decimal:
    return Decimal("{0:.6f}".format(d.timestamp()))



# -*- coding: utf-8 -*-
from typing import List, Optional

from eventbus.domain.eventbus import AbstractEventBus, AbstractEventHandler
from eventbus.domain.whitehead import TEvent


class EventBus(AbstractEventBus):
    def __init__(self, handlers: Optional[List[AbstractEventHandler]] = None):
        self._subscriptions: List[AbstractEventHandler] = handlers or []

    def subscribe(self, handler: AbstractEventHandler) -> None:
        if handler not in self._subscriptions:
            self._subscriptions.append(handler)

    def unsubscribe(self, handler: AbstractEventHandler) -> None:
        if handler in self._subscriptions:
            self._subscriptions.remove(handler)

    async def publish(self, events: List[TEvent]) -> None:
        for handler in self._subscriptions[:]:
            if handler.predicate(events):
                await handler.handler(events)

    def is_event_handlers_empty(self) -> bool:
        return len(self._subscriptions) == 0

    def clear_event_handlers(self) -> None:
        self._subscriptions.clear()



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["python-dateutil>=2.8.1", ]

setup_requirements = ["pytest-runner", "pyyaml>=5.3.1", "argh>=0.26.2"]

test_requirements = ["pytest>=3", ]

setup(
    author="Alexander A. Dyshev",
    author_email="adyshev@gmail.com",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords=[
        "async domain entity",
        "async event bus",
        "async aggregate root",
        "domain eventbus"
    ],
    install_requires=requirements,
    license="MIT license",
    include_package_data=True,
    name="domain-eventbus",
    packages=find_packages(include=["eventbus", "eventbus.*"], exclude=["docs"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/adyshev/eventbus",
    version="0.1.6",
    zip_safe=False,
)



# -*- coding: utf-8 -*-
from abc import ABCMeta
from uuid import UUID, uuid4
from typing import Any, Dict, List, Type, Union, TypeVar, Optional
from decimal import Decimal

from eventbus.domain.decorators import subclassevents
from eventbus.domain.eventbus import AbstractEventBus
from eventbus.domain.events import (
    EventWithOriginatorID, CreatedEvent, AttributeChangedEvent, DiscardedEvent,
    DomainEvent, EventWithOriginatorVersion, EventWithTimestamp
)
from eventbus.domain.exceptions import OriginatorIDError, EntityIsDiscarded, OriginatorVersionError
from eventbus.domain.versioning import Upcastable
from eventbus.domain.whitehead import EnduringObject
from eventbus.util.topic import get_topic, resolve_topic


class MetaDomainEntity(ABCMeta):
    __subclassevents__ = False

    def __init__(cls, name: str, *args: Any, **kwargs: Any) -> None:  # noqa
        super().__init__(name, *args, **kwargs)
        if cls.__subclassevents__ is True:
            subclassevents(cls)


TDomainEntity = TypeVar("TDomainEntity", bound="DomainEntity")
TDomainEvent = TypeVar("TDomainEvent", bound="DomainEntity.Event")


class DomainEntity(Upcastable, EnduringObject, metaclass=MetaDomainEntity):
    """
    Supertype for domain model entity.
    """

    __subclassevents__ = False

    class Event(EventWithOriginatorID[TDomainEntity]):
        """
        Supertype for events of domain model entities.
        """

        def __check_obj__(self, obj: TDomainEntity) -> None:
            """
            Checks state of obj before mutating.

            :param obj: Domain entity to be checked.

            :raises OriginatorIDError: if the originator_id is mismatched
            """
            # Assert ID matches originator ID.
            if obj.id != self.originator_id:
                raise OriginatorIDError(
                    "'{0}' not equal to event originator ID '{1}'" "".format(obj.id, self.originator_id)
                )

    @classmethod
    async def __create__(
        cls: Type[TDomainEntity],
        event_bus: AbstractEventBus,
        originator_id: Optional[UUID] = None,
        event_class: Optional[Type["DomainEntity.Created[TDomainEntity]"]] = None,
        **kwargs: Any,
    ) -> TDomainEntity:
        """
        Creates a new domain entity.

        Constructs a "created" event, constructs the entity object
        from the event, publishes the "created" event, and returns
        the new domain entity object.

        :param cls DomainEntity: Class of domain event
        :param originator_id: ID of the new domain entity (defaults to ``uuid4()``).
        :param event_class: Domain event class to be used for the "created" event.
        :param kwargs: Other named attribute values of the "created" event.
        :return: New domain entity object.
        :rtype: DomainEntity
        """

        if originator_id is None:
            originator_id = uuid4()

        if event_class is None:
            created_event_class: Type[DomainEntity.Created[TDomainEntity]] = cls.Created
        else:
            created_event_class = event_class

        event = created_event_class(
            event_bus=event_bus, originator_id=originator_id, originator_topic=get_topic(cls), **kwargs
        )

        obj = event.__mutate__(None)

        if obj is None:
            raise ValueError("{0} returned None".format(type(event).__mutate__.__qualname__))

        await obj.__publish__(event)
        return obj

    class Created(CreatedEvent[TDomainEntity], Event[TDomainEntity]):
        """
        Triggered when an entity is created.
        """

        def __init__(self, event_bus: AbstractEventBus, originator_topic: str, **kwargs: Any):
            super(DomainEntity.Created, self).__init__(
                event_bus=event_bus, originator_topic=originator_topic, **kwargs
            )

        @property
        def originator_topic(self) -> str:
            """
            Topic (a string) representing the class of the originating domain entity.

            :rtype: str
            """
            return self.__dict__["originator_topic"]

        def __mutate__(self, obj: Optional[TDomainEntity]) -> Optional[TDomainEntity]:
            """
            Constructs object from an entity class,
            which is obtained by resolving the originator topic,
            unless it is given as method argument ``entity_class``.

            entity_class: Class of domain entity to be constructed.
            """
            entity_class: Type[TDomainEntity] = resolve_topic(self.originator_topic)
            return entity_class(**self.__entity_kwargs__)

        @property
        def __entity_kwargs__(self) -> Dict[str, Any]:
            kwargs = self.__dict__.copy()
            kwargs["event_bus"] = kwargs.pop("event_bus")
            kwargs["id"] = kwargs.pop("originator_id")
            kwargs.pop("originator_topic", None)
            kwargs.pop("__event_topic__", None)
            kwargs.pop("__event_hash_method_name__", None)
            return kwargs

    def __init__(self, event_bus: AbstractEventBus, id: UUID, **kwargs):
        super().__init__()
        self._event_bus = event_bus
        self._id = id
        self.__is_discarded__ = False

    @property
    def id(self) -> UUID:
        """The immutable ID of the domain entity.

        This value is set using the ``originator_id`` of the
        "created" event constructed by ``__create__()``.

        An entity ID allows an instance to be
        referenced and distinguished from others, even
        though its state may change over time.

        This attribute has the normal "public" format for a Python object
        attribute name, because by definition all domain entities have an ID.
        """
        return self._id

    @property
    def event_bus(self) -> AbstractEventBus:
        return self._event_bus

    async def __change_attribute__(self: TDomainEntity, name: str, value: Any, **kwargs) -> None:
        """
        Changes named attribute with the given value,
        by triggering an AttributeChanged event.
        """
        event_class: Type["DomainEntity.AttributeChanged[TDomainEntity]"] = self.AttributeChanged
        await self.__trigger_event__(event_class=event_class, name=name, value=value, **kwargs)

    class AttributeChanged(Event[TDomainEntity], AttributeChangedEvent[TDomainEntity]):
        """
        Triggered when a named attribute is assigned a new value.
        """

        def __mutate__(self, obj: Optional[TDomainEntity]) -> Optional[TDomainEntity]:
            obj = super(DomainEntity.AttributeChanged, self).__mutate__(obj)
            setattr(obj, self.name, self.value)
            return obj

    async def __discard__(self: TDomainEntity, **kwargs) -> None:
        """
        Discards self, by triggering a Discarded event.
        """
        event_class: Type["DomainEntity.Discarded[TDomainEntity]"] = self.Discarded
        await self.__trigger_event__(event_class=event_class, **kwargs)

    class Discarded(DiscardedEvent[TDomainEntity], Event[TDomainEntity]):
        """
        Triggered when a DomainEntity is discarded.
        """

        def __mutate__(self, obj: Optional[TDomainEntity]) -> Optional[TDomainEntity]:
            obj = super(DomainEntity.Discarded, self).__mutate__(obj)
            if obj is not None:
                obj.__is_discarded__ = True
            return None

    def __assert_not_discarded__(self) -> None:
        """
        Asserts that this entity has not been discarded.

        Raises EntityIsDiscarded exception if entity has been discarded already.
        """
        if self.__is_discarded__:
            raise EntityIsDiscarded("Entity is discarded")

    async def __trigger_event__(self, event_class: Type[TDomainEvent], **kwargs: Any) -> None:
        """
        Constructs, applies, and publishes a domain event.
        """
        self.__assert_not_discarded__()
        event: TDomainEvent = event_class(originator_id=self.id, **kwargs)
        self.__mutate__(event)
        await self.__publish__(event)

    def __mutate__(self, event: TDomainEvent) -> None:
        """
        Mutates this entity with the given event.

        This method calls on the event object to mutate this
        entity, because the mutation behaviour of different types
        of events was usefully factored onto the event classes, and
        the event mutate() method is the most convenient way to
        defined behaviour in domain models.

        However, as an alternative to implementing the mutate()
        method on domain model events, this method can be extended
        with a method that is capable of mutating an entity for all
        the domain event classes introduced by the entity class.

        Similarly, this method can be overridden entirely in subclasses,
        so long as all of the mutation behaviour is implemented in the
        mutator function, including the mutation behaviour of the events
        defined on the library event classes that would no longer be invoked.

        However, if the entity class defines a mutator function, or if a
        separate mutator function is used, then it must be involved in
        the event sourced repository used to replay events, which by default
        knows nothing about the domain entity class. In practice, this
        means having a repository for each kind of entity, rather than
        the application just having one repository, with each repository
        having a mutator function that can project the entity events
        into an entity.
        """
        if not isinstance(event, DomainEntity.Event):
            raise ValueError("Given Event is not an instance of DomainEntity.Event")

        event.__mutate__(self)

    async def __publish__(self, events: Union[TDomainEvent, List[TDomainEvent]]) -> None:
        """
        Publishes given event for subscribers in the application.

        :param events: domain event or list of events
        """
        if isinstance(events, DomainEvent):
            events = [events]

        await self._event_bus.publish(events)

    def __eq__(self, other: object) -> bool:
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


TVersionedEntity = TypeVar("TVersionedEntity", bound="VersionedEntity")
TVersionedEvent = TypeVar("TVersionedEvent", bound="VersionedEntity.Event")


class VersionedEntity(DomainEntity):
    def __init__(self, __version__: int, **kwargs: Any):
        super().__init__(**kwargs)
        self.___version__: int = __version__

    @property
    def __version__(self) -> int:
        return self.___version__

    async def __trigger_event__(self, event_class: Type[TDomainEvent], **kwargs: Any) -> None:
        """
        Increments the version number when an event is triggered.

        The event carries the version number that the originator
        will have when the originator is mutated with this event.
        (The event's "originator" version isn't the version of the
        originator before the event was triggered, but represents
        the result of the work of incrementing the version, which
        is then set in the event as normal. The Created event has
        version 0, and a newly created instance is at version 0.
        The second event has originator version 1, and so will the
        originator when the second event has been applied.
        """
        # Do the work of incrementing the version number.
        next_version = self.__version__ + 1
        # Trigger an event with the result of this work.
        await super(VersionedEntity, self).__trigger_event__(
            event_class=event_class, originator_version=next_version, **kwargs
        )

    class Event(
        EventWithOriginatorVersion[TVersionedEntity], DomainEntity.Event[TVersionedEntity],
    ):
        """Supertype for events of versioned entities."""

        def __mutate__(self, obj: Optional[TVersionedEntity]) -> Optional[TVersionedEntity]:
            obj = super(VersionedEntity.Event, self).__mutate__(obj)
            if obj is not None:
                obj.___version__ = self.originator_version
            return obj

        def __check_obj__(self, obj: TVersionedEntity) -> None:
            """
            Extends superclass method by checking the event's
            originator version follows (1 +) this entity's version.
            """
            super(VersionedEntity.Event, self).__check_obj__(obj)
            # Assert the version sequence is correct.
            if self.originator_version != obj.__version__ + 1:
                raise OriginatorVersionError(
                    (
                        "Event takes entity to version {0}, "
                        "but entity is currently at version {1}. "
                        "Event type: '{2}', entity type: '{3}', entity ID: '{4}'"
                        "".format(
                            self.originator_version,
                            obj.__version__,
                            type(self).__name__,
                            type(obj).__name__,
                            obj._id,  # noqa
                        )
                    )
                )

    class Created(DomainEntity.Created[TVersionedEntity], Event[TVersionedEntity]):
        """Published when a VersionedEntity is created."""

        def __init__(self, originator_version: int = 0, *args: Any, **kwargs: Any):
            super(VersionedEntity.Created, self).__init__(originator_version=originator_version, *args, **kwargs)

        @property
        def __entity_kwargs__(self) -> Dict[str, Any]:
            # Get super property.
            kwargs = super(VersionedEntity.Created, self).__entity_kwargs__
            kwargs["__version__"] = kwargs.pop("originator_version")
            return kwargs

    class AttributeChanged(Event[TVersionedEntity], DomainEntity.AttributeChanged[TVersionedEntity]):
        """Published when a VersionedEntity is changed."""

    class Discarded(Event[TVersionedEntity], DomainEntity.Discarded[TVersionedEntity]):
        """Published when a VersionedEntity is discarded."""


TTimestampedEntity = TypeVar("TTimestampedEntity", bound="TimestampedEntity")


class TimestampedEntity(DomainEntity):
    def __init__(self, __created_on__: Decimal, **kwargs: Any):
        super(TimestampedEntity, self).__init__(**kwargs)
        self.___created_on__ = __created_on__
        self.___last_modified__ = __created_on__

    @property
    def __created_on__(self) -> Decimal:
        return self.___created_on__

    @property
    def __last_modified__(self) -> Decimal:
        return self.___last_modified__

    class Event(DomainEntity.Event[TTimestampedEntity], EventWithTimestamp[TTimestampedEntity]):
        """Supertype for events of timestamped entities."""

        def __mutate__(self, obj: Optional[TTimestampedEntity]) -> Optional[TTimestampedEntity]:
            """Updates 'obj' with values from self."""
            obj = super(TimestampedEntity.Event, self).__mutate__(obj)
            if obj is not None:
                obj.___last_modified__ = self.timestamp
            return obj

    class Created(DomainEntity.Created[TTimestampedEntity], Event[TTimestampedEntity]):
        """Published when a TimestampedEntity is created."""

        @property
        def __entity_kwargs__(self) -> Dict[str, Any]:
            # Get super property.
            kwargs = super(TimestampedEntity.Created, self).__entity_kwargs__
            kwargs["__created_on__"] = kwargs.pop("timestamp")
            return kwargs

    class AttributeChanged(Event[TTimestampedEntity], DomainEntity.AttributeChanged[TTimestampedEntity]):
        """Published when a TimestampedEntity is changed."""

    class Discarded(Event[TTimestampedEntity], DomainEntity.Discarded[TTimestampedEntity]):
        """Published when a TimestampedEntity is discarded."""


TTimestampedVersionedEntity = TypeVar("TTimestampedVersionedEntity", bound="TimestampedVersionedEntity")


class TimestampedVersionedEntity(TimestampedEntity, VersionedEntity):
    class Event(
        TimestampedEntity.Event[TTimestampedVersionedEntity], VersionedEntity.Event[TTimestampedVersionedEntity],
    ):
        """Supertype for events of timestamped, versioned entities."""

    class Created(
        TimestampedEntity.Created[TTimestampedVersionedEntity],
        VersionedEntity.Created,
        Event[TTimestampedVersionedEntity],
    ):
        """Published when a TimestampedVersionedEntity is created."""

    class AttributeChanged(
        Event[TTimestampedVersionedEntity],
        TimestampedEntity.AttributeChanged[TTimestampedVersionedEntity],
        VersionedEntity.AttributeChanged[TTimestampedVersionedEntity],
    ):
        """Published when a TimestampedVersionedEntity is created."""

    class Discarded(
        Event[TTimestampedVersionedEntity],
        TimestampedEntity.Discarded[TTimestampedVersionedEntity],
        VersionedEntity.Discarded[TTimestampedVersionedEntity],
    ):
        """Published when a TimestampedVersionedEntity is discarded."""
